# Copyright (C) 2024 Matthew Richardson

from spheronizator import functions as box
from spheronizator.mol2parser import mol2parser
import numpy as np
import re
import warnings

class voxelBuilder:

    def __init__(self, config=None):
        
        # Load configuration file
        self._get_config(config)

        # Attribute bondTypeDict determines the indexing of the output array in addition to the types of bonds that will be present in the output. Strings must match what is present in the associated mol2 file.
        self.bondTypeDict={
                '1':0,      # Single bonds
                '2':1,      # Double bonds
                '3':2,      # Triple bonds
                'am':3,     # Amides
                'ar':4      # Aromatics
                }

        # Attribute atomTypeDict determines the indexing of the output array in addition to the types of atoms that will be present in the output. Strings must match what is present in the first character of the associated mol2 file atom type. 
        self.atomTypeDict={
                'H':0,
                'C':1,
                'N':2,
                'O':3,
                'P':4,
                'S':5
                }

    def reloadConfig(self, configPath=None):
        
        self._get_config(configPath)

    def parse(self, pdbfile, mol2file=None):

        # Wrapper method for mol2parser class. This simplifies the interface.

        parser=mol2parser(pdbfile, mol2file)    # Create instance of parser object, parse and update atom objects
        self.structure=parser.structure
        self.residues=parser.residues

        # Only get atoms that belong to amino acids
        self.atoms=[atom for atom in parser.atoms if atom.isAA]

        self._get_resnames()
    
    def buildData(self):
        
        self._get_voxels()                      # Generate our voxels we will need for each box and store as object attribute
        self._init_arrays()                     # Initialize all arrays we will need for output data

        # Sanity check
        if self.voxels.shape[0:3]!=self.output.shape[1:4]:
            raise ValueError("Voxel array shape and output array shape do not match! File a bug report.")

        
        for i in range(len(self.residues)):
    
            # Build data for atom abscence / presence
            foundAtomIndices, projectedCoords=self._build_box(self.residues[i])
            self._process_box(foundAtomIndices, projectedCoords, i)

            # Build data for bond information
            self._process_box_bonds(foundAtomIndices, i)
               
    def check_collision(self):

        return np.any(self.output > 1)

    def find_collisions(self):

        return np.transpose(np.where(self.output > 1))
    
    def _get_config(self, configPath=None):

        if configPath is None:
            configPath='config'

        try:
           with open(configPath, 'r') as configFile:
                
                config=[]
                delimiter=re.compile(r'=')

                for line in configFile:
                    
                    line=line.strip()

                    if line and not line.startswith('#'):
                        config.append(delimiter.split(line))
                
                
                self.config=dict(config)

        except:

            self.config={
                        'boxSize':20,
                        'voxelSpacing':1,
                        'useFloatVoxels':'True',
                        'dataType':'bool',
                        'useWarnings':'True',
                        'useSpheres':'False'
                    }

        # Unpack values to attributes. This allows this values to be changed after initialization.
        self.boxSize            = int(  self.config['boxSize'])
        self.voxelSpacing       = int(  self.config['voxelSpacing'])
        self.useFloatVoxels     =       self.config['useFloatVoxels'].startswith('True')
        self.dataType           =       self.config['dataType']
        self.useWarnings        =       self.config['useWarnings'].startswith('True')
        self.useSpheres         =       self.config['useSpheres'].startswith('True')
    
    def _init_arrays(self):
        
        voxelArrayLength=np.rint(np.divide(self.boxSize, self.voxelSpacing) + 1).astype(int)
        residueCount=len(self.residues)

        # Output array for atom presence / abscence
        self.output=np.zeros((
                residueCount,                   # Number of residues in protein
                voxelArrayLength,               # Size of voxel array
                voxelArrayLength,
                voxelArrayLength,
                len(self.atomTypeDict),         # Size of atomTypeDict representing the count of atom channels
                2                               # Last dimension indicates whether or not atom belongs to the parent residue of the box
                ), dtype=self.dataType)

        self.outputBonds=np.zeros((
                residueCount,
                len(self.bondTypeDict)          # Size of bondTypeDict representing the count of different types of bonds to be considered
                ), dtype=int)
    
    def _get_resnames(self):

        self.resnames=[residue.get_resname() for residue in self.residues]
    
    def _build_box(self, residue):

        # Obtain coordinate projection of all atom objects about a standard position based on the parent residue
        projectedAtoms=box.get_boxProjection(residue, self.atoms)

        # Get the indices of all atoms contained within the box
        if self.useSpheres:
            foundAtomIndices=box.buildSphere(projectedAtoms, self.boxSize)
        else:
            foundAtomIndices=box.buildBox(projectedAtoms, self.boxSize)

        return foundAtomIndices, projectedAtoms
                
    def _get_voxels(self):

        # Generate voxels and store as attribute.  Voxels are not unique to each box, so they only need to be generated once.
        if self.useFloatVoxels:
            self.voxels=box.get_floatVoxels(self.boxSize, self.voxelSpacing)

        else:
            self.voxels=box.get_voxels(self.boxSize, self.voxelSpacing)

    def _process_box(self, foundAtomIndices, projectedCoords, residueIndex):  

        for i in foundAtomIndices:
            
            atom=self.atoms[i]
             
            try:
                atomTypeIndex=self.atomTypeDict[atom.atomType]
            except:
                continue
 
            voxelIndex, voxelCoords=box.get_closestVoxel(projectedCoords[i], self.voxels)

            if self.useWarnings and self.output[residueIndex][voxelIndex][atomTypeIndex][0]:
                voxelStr=str(voxelIndex).strip(r"()")
                print(f"Warning: Atom collision at index ({residueIndex}, {voxelStr}, {atomTypeIndex}) by atom with index {i}.")
            
            self.output[residueIndex][voxelIndex][atomTypeIndex][0]+=1

            if atom.isAA and atom.residueIndex==residueIndex:
                self.output[residueIndex][voxelIndex][atomTypeIndex][1]+=1

    def _process_box_bonds(self, foundAtomIndices, residueIndex):
        
        mol2indices=[self.atoms[i].mol2atomIndex for i in foundAtomIndices]
        
        for i in foundAtomIndices:
            
            atom=self.atoms[i]

            if hasattr(atom, 'bondData'):
                for bond in atom.bondData:
                    if bond[0] in mol2indices:
                        try:
                            bondIndex=self.bondTypeDict[bond[1]]

                        except:
                            continue

                        self.outputBonds[residueIndex][bondIndex]+=1

    
    def _debug_export_boxes(self):

       # This function is used to export generated boxes into a directory as PDB files themselves for testing/debugging purposes. 
        
        from Bio.PDB.PDBIO import Select
        from Bio.PDB import PDBIO

        io=PDBIO()
        
        class boxSelect(Select):
            def accept_atom(self, atom):
                if atom in foundAtoms:
                    return True
                else:
                    return False

        for index,residue in enumerate(self.residues):

            foundAtomsIndex, projectedCoords=self._build_box(residue)
            foundAtoms=[self.atoms[i] for i in foundAtomsIndex]
            filename_out="output/box" + str(index+1) + ".pdb"
            io.set_structure(self.structure)
            io.save(filename_out,boxSelect())
            print(filename_out)




# Copyright (C) 2024 Matthew Richardson

'''
The goal of this parser is to maintain interoperability with Biopython's parsing of PDB files while adding information within mol2 files to Biopython structure objects. Although the name of this class is 'mol2parser', it is a combination PDB parser and mol2parser. The private methods may be used to parse mol2 files individually, but this parser is not designed to be used this way.

'''

import numpy as np
import re
from Bio.PDB.PDBParser import PDBParser
from Bio.PDB.Polypeptide import is_aa

class mol2parser:
    
    def __init__(self,pdbfile,mol2file=None):
        
        if mol2file is None:
            mol2file=pdbfile + '.mol2'

        try:
            self._parse_pdb(pdbfile)                # First parse the PDB file to get a structure object
            self._parse_mol2(mol2file)              # Parse corresponding mol2file to extract additional data about the protein
            self._update_records()                  # Add the mol2file data to the atom objects obtained from the PDB file

        except ParseError:
            print("Files were unable to be parsed")

    def _update_records(self):

        # Update the atom objects extracted from the PDB file with information from the corresponding mol2 file

        try:
            self._add_detailed_atom_type()
            self._add_atom_type()
            self._add_bond_info()
            self._add_atom_index()
        
        except ParseError:
            print("Unable to update atomic records from mol2 file")

    def _parse_pdb(self, file):

        # Call instance of Biopython PDB parser to get structure, residues, and atom objects
        
        pdb=PDBParser()
        self.structure=pdb.get_structure(file, file)
        
        self.atoms=[]
        self.residues=[]
        residueIndex=-1

        for residue in self.structure.get_residues():

            if is_aa(residue):

                self.residues.append(residue)
                residueIndex+=1

                for atom in residue.get_atoms():
                    atom.isAA=True
                    atom.residueIndex=residueIndex
                    self.atoms.append(atom)

            else:

                for atom in residue.get_atoms():
                    atom.isAA=False
                    atom.residueIndex=False
                    self.atoms.append(atom)

    def _parse_mol2(self, file):
        
        with open(file, 'r') as mol2file:
            
            sectionCounter=-1
            self.sections=[]
            self.parsedData=[]
            delimiter=re.compile(r'\s+')
            
            for line in mol2file:

                line=line.strip()
                
                if line.startswith('@<TRIPOS>'):
                    sectionCounter+=1
                    self.parsedData.append([line.removeprefix('@<TRIPOS>')])
                    sectionName=self.parsedData[sectionCounter][0]
                    self.sections.append([sectionName,sectionCounter])
            
                elif line:
                    line=delimiter.split(line)      
                    self.parsedData[sectionCounter].append(line)

            self.sections=dict(self.sections)
                            
                    
    def _get_atoms(self):

        return self.parsedData[self.sections['ATOM']][1:]

    def _get_bonds(self):

        return self.parsedData[self.sections['BOND']][1:]

    def _add_detailed_atom_type(self):
        
        mol2atoms=self._get_atoms()

        for i in range(len(self.atoms)):
            self.atoms[i].detailedAtomType=mol2atoms[i][5]

    def _add_atom_type(self):

        for i in range(len(self.atoms)):
            self.atoms[i].atomType=self.atoms[i].detailedAtomType[0]
        
    def _add_atom_index(self):

        for i in range(len(self.atoms)):
            self.atoms[i].mol2atomIndex=i
    
    def _add_bond_info(self):
        
        mol2bonds=self._get_bonds()

        for bond in mol2bonds:
            
            originID, targetID, bondType=int(bond[1]), int(bond[2]), bond[3]
            originIndex, targetIndex=originID-1, targetID-1

            # Sanity Check
            if not originID==self.atoms[originIndex].get_serial_number():
                raise ValueError("mol2 atom index does not match PDB index")

            # Test to see if the atom object already contains the bondData attribute
            if not hasattr(self.atoms[originIndex],'bondData'):
                self.atoms[originIndex].bondData=[]

            self.atoms[originIndex].bondData.append([targetIndex, bondType])

class ParseError(Exception):
    pass



# Copyright (C) 2024 Matthew Richardson

import numpy as np
import math
from Bio.PDB.PDBParser import PDBParser
from Bio.PDB import vectors

def get_structureLimits(atomCoords):
    ''' 
    Get the upper and lower bounds for every dimension from the given set of coordinates.
    Used to get the ranges of cartesian coordiantes that describe the entire structure.
    
    Output array is in the following format:
            x y z
        min . . .
        max . . .
        
    '''
    
    structureLimits=np.array([
        [np.min(atomCoords[:,dimension]) for dimension in range(0,3)], 
        [np.max(atomCoords[:,dimension]) for dimension in range(0,3)]],
        dtype=np.float32
    )
    
    return structureLimits

def get_samplePoints(structureLimits,spacing=10):
    '''
    Given boundaries, generate a set of sample points spaced at regular intervals.
    Currently, we round to the nearest integer.
    
    Default spacing interval is 10 angstrom

    Output array is in the following format:

            x y z
    point 1 . . .
    point 2 . . .
    
    '''
    
    if not isinstance(spacing, int):
        raise ValueError("sampleRate is not a integer value!")

    # Round min and max to the nearest integer
    structureLimits[0]=np.floor(structureLimits[0])
    structureLimits[1]=np.ceil(structureLimits[1])
    
    # Generate an array of sample points for each dimension
    # Add one (1) to the end to include end-point
    x_points=np.arange(structureLimits[0,0],structureLimits[1,0]+1,spacing, dtype=int)
    y_points=np.arange(structureLimits[0,1],structureLimits[1,1]+1,spacing, dtype=int)
    z_points=np.arange(structureLimits[0,2],structureLimits[1,2]+1,spacing, dtype=int)

    # Verify ranges for debugging purposes
    if (x_points[-1]<structureLimits[1,0] or
        y_points[-1]<structureLimits[1,1] or
        z_points[-1]<structureLimits[1,2]):
        raise ValueError("One of the atoms is out of range!")

    '''
    Use numpy.meshgrid to generate the sample points.
    
    Equivalent to the following:
    
    samplePoints=[]
    for x in x_points
        for y in y_points
            for z in z_points
                samplePoints.append([x,y,z])
    '''
    
    x,y,z=np.meshgrid(x_points, y_points, z_points, indexing='ij')
    samplePoints=np.stack((x, y, z),axis=-1).reshape(-1,3)
    
    return samplePoints

def get_voxels(boxSize=20,spacing=1):

    '''
    Return a 3D array of regularly spaced sample points to serve as voxels for our model.

    '''
    
    if not isinstance(spacing, int):
        raise ValueError("spacing is not a integer value!")

    if not isinstance(boxSize, int):
        raise ValueError("boxSize is not a integer value!")

    if boxSize%2!=0:
        raise ValueError("boxSize is not divisible by 2!")

    delta=np.intc(boxSize>>1)
    
    x_points=np.arange(-delta, delta+spacing, spacing, dtype=int)
    y_points=np.arange(-delta, delta+spacing, spacing, dtype=int)
    z_points=np.arange(-delta, delta+spacing, spacing, dtype=int)

    x,y,z=np.meshgrid(x_points, y_points, z_points, indexing='ij')
    samplePoints=np.stack((x, y, z),axis=-1)

    return samplePoints

def get_floatVoxels(boxSize=20,spacing=1):

    '''
    Return a 3D array of regularly spaced sample points to serve as voxels for our model. Function is able to output voxels as floats,
    which enables the support of voxelSpacings less than 1. Supercedes above function, since calculations are all floating point anyways,
    and there is no performance loss.

    '''

    if not (np.isclose(np.mod(boxSize, spacing), 0, 1e-9) or
            np.isclose(np.mod(boxSize, spacing), spacing, 1e-9)):
        raise ValueError("spacing does not divide boxSize!")

    steps=np.rint(np.divide(boxSize, spacing) + 1).astype(int)
    delta=np.divide(boxSize, 2)
 
    x_points=np.linspace(-delta, delta, steps, endpoint=True)
    y_points=np.linspace(-delta, delta, steps, endpoint=True)
    z_points=np.linspace(-delta, delta, steps, endpoint=True)

    x,y,z=np.meshgrid(x_points, y_points, z_points, indexing='ij')
    samplePoints=np.stack((x, y, z),axis=-1)

    return samplePoints

def get_closestVoxel(point,voxels):

    '''
    Given a point and array of voxels, return the index of the voxel that lies closest to that point.

    '''
    # Compute the distance from every voxel to the specified point
    computedDistances=np.linalg.norm(voxels-point, axis=-1)

    # Find the smallest distance and return the 3D index of this point, not the flattened index
    minIndex=np.unravel_index(np.argmin(computedDistances), computedDistances.shape)

    return minIndex,voxels[minIndex]

def get_centralAA(samplePoints,atomCoords,boxSize=20):
    for sample in samplePoints:
        foundAtoms=buildBox(sample,atomCoords,boxSize)
        if len(foundAtoms)!=0:
            closestAtom=foundAtoms[0]
            x=np.subtract(atomCoords[closestAtom],sample)
            shortestDistance=np.dot(x,x)

            for atom in foundAtoms:
                x=np.subtract(atomCoords[atom],sample)
                d=np.dot(x,x) # We don't need the real distance, just d^2
                if d<shortestDistance:
                    shortestDistance=d
                    closestAtom=atom
    
def get_scanRange(origin,delta,size,voxel):
    scanRange=np.zeros((3,size),dtype=int)
    for dimension,component in enumerate(boxOrigin):
        # Add one (1) voxel to maximum to include endpoint
        scanMin,scanMax=component-delta,component+delta+voxel
        scanRange[dimension,:]=np.arange(scanMin,scanMax,voxel)
    return scanRange

def get_boxProjection(residue,atoms):
    '''
    Given a Biopython residue object to act as the center, and array of Biopython atom objects,
    return a new array of atom coordinates that have been projected onto a standard position.

    We may want to implement this as a matrix projection.
    
    Define new coordinate system as follows:
    Let N, CA, C be coplanar and parallel to the xy-plane.
    Let the vector i=N-CA be parallel to the x-axis.
    Then the vector k=ix(C-CA) is parallel to the z-axis,
    and the vector j=kxi is parallel to the y-axis.
    
    '''

    # Extract three vectors from the residue
    n=residue["N"].get_vector()
    ca=residue["CA"].get_vector()
    c=residue["C"].get_vector()

    # Create two vectors, i and v, in the xy-plane with the alpha carbon as the origin.
    i=n-ca
    v=c-ca
    
    # Create vector in the direction of the z-axis
    k=i**v # Cross Product

    # Create vector in the direction of the y-axis
    j=k**i

    # Make each a unit vector. This saves us from having to recompute later
    i=i/np.linalg.norm(i)
    j=j/np.linalg.norm(j)
    k=k/np.linalg.norm(k)

    # Sanity-check: See that i,j,k are orthagonal to each other
    if not (math.isclose(i*j, 0, abs_tol=1e-6) or
            math.isclose(i*k, 0, abs_tol=1e-6) or
            math.isclose(j*k, 0, abs_tol=1e-6)):
            raise ValueError("Something is wrong with our imposed coordinate system!")

    projectedCoordinates=[]
    for atom in atoms:
        u=atom.get_vector()-ca
        x,y,z=i*u,j*u,k*u # Dot product
        projection=np.array([x, y, z])

        # Sanity-check: distances to central atom should be the same in both coordinate systems
        if not math.isclose(np.linalg.norm(u),np.linalg.norm(projection), abs_tol=1e-6):
            raise ValueError("Something is wrong with our imposed coordinate system!")

        projectedCoordinates.append(projection)
        
    return projectedCoordinates

def buildBox(atomCoords, boxSize=20):

    delta=np.divide(boxSize, 2)
    boxOrigin=(0,0,0)

    # Get our scanning range for the algorithm from the origin point
    # We want to avoid recomputing this range every iteration because it is constant
    scanMin=[component-delta for component in boxOrigin]
    scanMax=[component+delta+1 for component in boxOrigin] # Add 1 to include endpoint

    foundAtoms=[]
    for index,atom in enumerate(atomCoords):
        if np.array_equal(atom,boxOrigin):
            continue # Exclude the central atom from atoms to search
        for dimension,component in enumerate(atom[:3]):
            if (component-scanMin[dimension])*(component-scanMax[dimension]) > 0:
                break # One dimension was out of range, so no need to check the rest
        else:
            # We have found an atom that is within the box
            foundAtoms.append(index)

    return foundAtoms

def buildSphere(atomCoords, boxSize=20):
    
    radius=np.divide(boxSize, 2)
    computedDistances=np.linalg.norm(atomCoords, axis=-1)
    foundAtomIndices=np.where(computedDistances <= radius)[0]
    
    return foundAtomIndices
            
